@@ -8,6 +8,7 @@ import (
 	"context"
 	"crypto/tls"
 	"errors"
+	"fmt"
 	"math/rand"
 	"net"
 	"sort"
@@ -47,6 +48,55 @@ func (o byPriority) Less(i, j int) bool {
 	return o[i].Priority < o[j].Priority
 }
 
+func getCNAME(server string) (string, error) {
+	// Lookup CNAME
+	cname, err := net.LookupCNAME(server)
+	return cname, err
+}
+
+func getCNAMEDoT(server string, resolver string) (string, error) {
+	var cname string
+	var err error
+	var d tls.Dialer
+	s := strings.Split(resolver, "#")
+	if len(s) != 2 {
+		return cname, errors.New("wrong DoT server syntax")
+	}
+	ip := net.ParseIP(s[0])
+	if ip == nil {
+		return cname, errors.New("invalid resolver IP")
+	}
+	dnsServer := net.JoinHostPort(s[0], "853")
+	d.Config = &tls.Config{
+		ServerName: s[1],
+	}
+	r := &net.Resolver{
+		Dial: func(cont context.Context, network, address string) (net.Conn, error) {
+			return d.DialContext(cont, "tcp", dnsServer)
+		},
+	}
+	cname, err = r.LookupCNAME(context.Background(), server)
+	return cname, err
+}
+
+func getCNAMECustomResolver(server string, resolver string) (string, error) {
+	var cname string
+	var err error
+	var d net.Dialer
+	ip := net.ParseIP(resolver)
+	if ip == nil {
+		return cname, errors.New("invalid resolver IP")
+	}
+	dnsServer := net.JoinHostPort(resolver, "53")
+	r := &net.Resolver{
+		Dial: func(cont context.Context, network, address string) (net.Conn, error) {
+			return d.DialContext(cont, "udp", dnsServer)
+		},
+	}
+	cname, err = r.LookupCNAME(context.Background(), server)
+	return cname, err
+}
+
 func getSRVDoT(server string, srvType string,
 	resolver string) ([]SRV, error) {
 	var records []SRV
@@ -56,6 +106,10 @@ func getSRVDoT(server string, srvType string,
 	if len(s) != 2 {
 		return records, errors.New("wrong DoT server syntax")
 	}
+	ip := net.ParseIP(s[0])
+	if ip == nil {
+		return records, errors.New("invalid resolver IP")
+	}
 	dnsServer := net.JoinHostPort(s[0], "853")
 	d.Config = &tls.Config{
 		ServerName: s[1],
@@ -84,6 +138,10 @@ func getSRVCustomResolver(server string, srvType string,
 	var records []SRV
 	var err error
 	var d net.Dialer
+	ip := net.ParseIP(resolver)
+	if ip == nil {
+		return records, errors.New("invalid resolver IP")
+	}
 	dnsServer := net.JoinHostPort(resolver, "53")
 	r := &net.Resolver{
 		Dial: func(cont context.Context, network, address string) (net.Conn, error) {
@@ -125,6 +183,11 @@ func getSRV(server string, srvType string) ([]SRV, error) {
 
 // LookupXmppServer returns the xmpp-server SRV records.
 func LookupXmppServer(server string) ([]SRV, error) {
+	// Look up CNAME
+	cname, err := getCNAME(server)
+	if err == nil {
+		server = cname
+	}
 	// Look up xmpp-server SRV
 	records, err := getSRV(server, "xmpp-server")
 	return records, err
@@ -132,15 +195,28 @@ func LookupXmppServer(server string) ([]SRV, error) {
 
 // LookupXmppServer returns the xmpp-server SRV records.
 func (c *Config) LookupXmppServer(server string) ([]SRV, error) {
+	var cname string
 	var records []SRV
 	var err error
 	// Look up xmpp-server SRV
 	switch {
 	case c.Resolver == "":
+		cname, err = getCNAME(server)
+		if err == nil {
+			server = cname
+		}
 		records, err = getSRV(server, "xmpp-server")
 	case c.Resolver != "" && !c.DoT:
+		cname, err = getCNAMECustomResolver(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		records, err = getSRVCustomResolver(server, "xmpp-server", c.Resolver)
 	case c.Resolver != "" && c.DoT:
+		cname, err = getCNAMEDoT(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		records, err = getSRVDoT(server, "xmpp-server", c.Resolver)
 	default:
 		return records, errors.New("xmppsrv: invalid resolver setting")
@@ -151,24 +227,62 @@ func (c *Config) LookupXmppServer(server string) ([]SRV, error) {
 
 // LookupXmppsServer returns the xmpps-server SRV records.
 func LookupXmppsServer(server string) ([]SRV, error) {
+	// Look up CNAME
+	cname, err := getCNAME(server)
+	if err == nil {
+		server = cname
+	}
 	// Look up xmpps-server SRV
 	records, err := getSRV(server, "xmpps-server")
 	return records, err
 }
 
+// LookupCNAME returns the final target name.
+func (c *Config) LookupCNAME(server string) (string, error) {
+	var cname string
+	var err error
+	switch {
+	case c.Resolver == "":
+		// Look up CNAME
+		cname, err = getCNAME(server)
+	case c.Resolver != "" && !c.DoT:
+		// Look up CNAME
+		cname, err = getCNAMECustomResolver(server, c.Resolver)
+	case c.Resolver != "" && c.DoT:
+		// Look up CNMAE
+		cname, err = getCNAMEDoT(server, c.Resolver)
+	default:
+		return cname, errors.New("xmppsrv: invalid resolver setting")
+	}
+	return cname, err
+}
+
 // LookupXmppsServer returns the xmpps-server SRV records.
 func (c *Config) LookupXmppsServer(server string) ([]SRV, error) {
+	var cname string
 	var records []SRV
 	var err error
 	switch {
 	case c.Resolver == "":
+		cname, err = getCNAME(server)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpps-server SRV
 		records, err = getSRV(server, "xmpps-server")
 	case c.Resolver != "" && !c.DoT:
+		cname, err = getCNAMECustomResolver(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpps-server SRV
 		records, err = getSRVCustomResolver(server, "xmpps-server",
 			c.Resolver)
 	case c.Resolver != "" && c.DoT:
+		cname, err = getCNAMEDoT(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpps-server SRV
 		records, err = getSRVDoT(server, "xmpps-server", c.Resolver)
 	default:
@@ -179,6 +293,11 @@ func (c *Config) LookupXmppsServer(server string) ([]SRV, error) {
 
 // LookupXmppClient returns the xmpp-server SRV records.
 func LookupXmppClient(server string) ([]SRV, error) {
+	// Look up CNAME
+	cname, err := getCNAME(server)
+	if err == nil {
+		server = cname
+	}
 	// Look up xmpp-client SRV
 	records, err := getSRV(server, "xmpp-client")
 	return records, err
@@ -186,17 +305,30 @@ func LookupXmppClient(server string) ([]SRV, error) {
 
 // LookupXmppClient returns the xmpp-server SRV records.
 func (c *Config) LookupXmppClient(server string) ([]SRV, error) {
+	var cname string
 	var records []SRV
 	var err error
 	switch {
 	case c.Resolver == "":
+		cname, err = getCNAME(server)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpp-client SRV
 		records, err = getSRV(server, "xmpp-client")
 	case c.Resolver != "" && !c.DoT:
+		cname, err = getCNAMECustomResolver(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpp-client SRV
 		records, err = getSRVCustomResolver(server, "xmpp-client",
 			c.Resolver)
 	case c.Resolver != "" && c.DoT:
+		cname, err = getCNAMEDoT(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpp-client SRV
 		records, err = getSRVDoT(server, "xmpp-client", c.Resolver)
 	default:
@@ -207,6 +339,11 @@ func (c *Config) LookupXmppClient(server string) ([]SRV, error) {
 
 // LookupXmppsClient returns the xmpp-server SRV records.
 func LookupXmppsClient(server string) ([]SRV, error) {
+	// Look up CNAME
+	cname, err := getCNAME(server)
+	if err == nil {
+		server = cname
+	}
 	// Look up xmpps-client SRV
 	records, err := getSRV(server, "xmpps-client")
 	return records, err
@@ -214,17 +351,30 @@ func LookupXmppsClient(server string) ([]SRV, error) {
 
 // LookupXmppsClient returns the xmpp-server SRV records.
 func (c *Config) LookupXmppsClient(server string) ([]SRV, error) {
+	var cname string
 	var records []SRV
 	var err error
 	switch {
 	case c.Resolver == "":
+		cname, err = getCNAME(server)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpps-client SRV
 		records, err = getSRV(server, "xmpps-client")
 	case c.Resolver != "" && !c.DoT:
+		cname, err = getCNAMECustomResolver(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpps-client SRV
 		records, err = getSRVCustomResolver(server, "xmpps-client",
 			c.Resolver)
 	case c.Resolver != "" && c.DoT:
+		cname, err = getCNAMEDoT(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpps-client SRV
 		records, err = getSRVDoT(server, "xmpps-client", c.Resolver)
 	default:
@@ -237,12 +387,17 @@ func (c *Config) LookupXmppsClient(server string) ([]SRV, error) {
 // priority and weight.
 func LookupClient(server string) ([]SRV, error) {
 	var records, records2 []SRV
-	var err error
+	var err, err2 error
 	rand.Seed(time.Now().UnixNano())
+	// Look up CNAME
+	cname, err := getCNAME(server)
+	if err == nil {
+		server = cname
+	}
 	// Look up xmpp-client SRV
-	records, _ = getSRV(server, "xmpp-client")
+	records, err = getSRV(server, "xmpp-client")
 	// Look up xmpps-client SRV records.
-	records2, _ = getSRV(server, "xmpps-client")
+	records2, err2 = getSRV(server, "xmpps-client")
 	switch rand.Intn(2) {
 	case 0:
 		records = append(records, records2...)
@@ -251,6 +406,15 @@ func LookupClient(server string) ([]SRV, error) {
 	default:
 		records = append(records, records2...)
 	}
+	switch {
+	case err != nil && err2 != nil:
+		return records, fmt.Errorf("failure in xmpp-client lookup: %v\n"+
+			"failure in xmpps-client lookup: %v", err, err2)
+	case err != nil:
+		return records, err
+	case err2 != nil:
+		return records, err2
+	}
 	switch len(records) {
 	case 0:
 		return records, errors.New("no client records found")
@@ -267,25 +431,38 @@ func LookupClient(server string) ([]SRV, error) {
 // LookupClient returns the xmpp-client and xmpps-client SRV records sorted by
 // priority and weight.
 func (c *Config) LookupClient(server string) ([]SRV, error) {
+	var cname string
 	var records, records2 []SRV
-	var err error
+	var err, err2 error
 	rand.Seed(time.Now().UnixNano())
 	switch {
 	case c.Resolver == "":
+		cname, err = getCNAME(server)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpp-client SRV
-		records, _ = getSRV(server, "xmpp-client")
+		records, err = getSRV(server, "xmpp-client")
 		// Look up xmpps-client SRV records.
-		records2, _ = getSRV(server, "xmpps-client")
+		records2, err2 = getSRV(server, "xmpps-client")
 	case c.Resolver != "" && !c.DoT:
+		cname, err = getCNAMECustomResolver(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpp-client SRV
-		records, _ = getSRVCustomResolver(server, "xmpp-client", c.Resolver)
+		records, err = getSRVCustomResolver(server, "xmpp-client", c.Resolver)
 		// Look up xmpps-client SRV records.
-		records2, _ = getSRVCustomResolver(server, "xmpps-client", c.Resolver)
+		records2, err2 = getSRVCustomResolver(server, "xmpps-client", c.Resolver)
 	case c.Resolver != "" && c.DoT:
+		cname, err = getCNAMEDoT(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpp-client SRV
-		records, _ = getSRVDoT(server, "xmpp-client", c.Resolver)
+		records, err = getSRVDoT(server, "xmpp-client", c.Resolver)
 		// Look up xmpps-client SRV records.
-		records2, _ = getSRVDoT(server, "xmpps-client", c.Resolver)
+		records2, err2 = getSRVDoT(server, "xmpps-client", c.Resolver)
 	default:
 		return records, errors.New("xmppsrv: invalid resolver setting")
 	}
@@ -297,6 +474,15 @@ func (c *Config) LookupClient(server string) ([]SRV, error) {
 	default:
 		records = append(records, records2...)
 	}
+	switch {
+	case err != nil && err2 != nil:
+		return records, fmt.Errorf("failure in xmpp-client lookup: %v\n"+
+			"failure in xmpps-client lookup: %v", err, err2)
+	case err != nil:
+		return records, err
+	case err2 != nil:
+		return records, err2
+	}
 	switch len(records) {
 	case 0:
 		return records, errors.New("no client records found")
@@ -313,12 +499,16 @@ func (c *Config) LookupClient(server string) ([]SRV, error) {
 // LookupServer returns the xmpp-server and xmpps-server SRV records sorted by
 // priority and weight.
 func LookupServer(server string) ([]SRV, error) {
-	var err error
+	var err, err2 error
 	rand.Seed(time.Now().UnixNano())
+	cname, err := getCNAME(server)
+	if err == nil {
+		server = cname
+	}
 	// Look up xmpp-client SRV
-	records, _ := getSRV(server, "xmpp-server")
+	records, err := getSRV(server, "xmpp-server")
 	// Look up xmpps-client SRV records.
-	records2, _ := getSRV(server, "xmpps-server")
+	records2, err2 := getSRV(server, "xmpps-server")
 	switch rand.Intn(2) {
 	case 0:
 		records = append(records, records2...)
@@ -327,6 +517,15 @@ func LookupServer(server string) ([]SRV, error) {
 	default:
 		records = append(records, records2...)
 	}
+	switch {
+	case err != nil && err2 != nil:
+		return records, fmt.Errorf("failure in xmpp-server lookup: %v\n"+
+			"failure in xmpps-server lookup: %v", err, err2)
+	case err != nil:
+		return records, err
+	case err2 != nil:
+		return records, err2
+	}
 	switch len(records) {
 	case 0:
 		return records, errors.New("no server records found")
@@ -343,27 +542,40 @@ func LookupServer(server string) ([]SRV, error) {
 // LookupServer returns the xmpp-server and xmpps-server SRV records sorted by
 // priority and weight.
 func (c *Config) LookupServer(server string) ([]SRV, error) {
+	var cname string
 	var records, records2 []SRV
-	var err error
+	var err, err2 error
 	rand.Seed(time.Now().UnixNano())
 	switch {
 	case c.Resolver == "":
+		cname, err = getCNAME(server)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpp-client SRV
-		records, _ = getSRV(server, "xmpp-server")
+		records, err = getSRV(server, "xmpp-server")
 		// Look up xmpps-client SRV records.
-		records2, _ = getSRV(server, "xmpps-server")
+		records2, err2 = getSRV(server, "xmpps-server")
 	case c.Resolver != "" && !c.DoT:
+		cname, err = getCNAMECustomResolver(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpp-client SRV
-		records, _ = getSRVCustomResolver(server, "xmpp-server",
+		records, err = getSRVCustomResolver(server, "xmpp-server",
 			c.Resolver)
 		// Look up xmpps-client SRV records.
-		records2, _ = getSRVCustomResolver(server, "xmpps-server",
+		records2, err2 = getSRVCustomResolver(server, "xmpps-server",
 			c.Resolver)
 	case c.Resolver != "" && c.DoT:
+		cname, err = getCNAMEDoT(server, c.Resolver)
+		if err == nil {
+			server = cname
+		}
 		// Look up xmpp-client SRV
-		records, _ = getSRVDoT(server, "xmpp-server", c.Resolver)
+		records, err = getSRVDoT(server, "xmpp-server", c.Resolver)
 		// Look up xmpps-client SRV records.
-		records2, _ = getSRVDoT(server, "xmpps-server", c.Resolver)
+		records2, err2 = getSRVDoT(server, "xmpps-server", c.Resolver)
 	default:
 		return records, errors.New("xmppsrv: invalid resolver setting")
 	}
@@ -375,6 +587,15 @@ func (c *Config) LookupServer(server string) ([]SRV, error) {
 	default:
 		records = append(records, records2...)
 	}
+	switch {
+	case err != nil && err2 != nil:
+		return records, fmt.Errorf("failure in xmpp-server lookup: %v\n"+
+			"failure in xmpps-server lookup: %v", err, err2)
+	case err != nil:
+		return records, err
+	case err2 != nil:
+		return records, err2
+	}
 	switch len(records) {
 	case 0:
 		return records, errors.New("no server records found")
